# coding=utf8
"""REST

Extends Service to allow for launching as an http accesible rest
service
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-17"

# Limit exports
__all__ = [ 'bottle', 'REST' ]

# Ouroboros imports
from jobject import jobject
import jsonb
import memory
from tools import clone

# Python imports
from datetime import datetime
import re
import sys
import traceback
from typing import List, Literal

# Pip imports
import bottle

# Local imports
from . import errors, response, service

class _Route(object):
	"""Route

	A private callable class used to store rest routes accessed by bottle
	"""

	__content_type = re.compile(r'^application\/json; charset=utf-?8$')
	"""Valid Content-Type"""

	__cors = None
	"""CORs regular expression"""

	__key_to_errors = {
		'data': errors.SERVICE_NO_DATA,
		'session': errors.SERVICE_NO_SESSION
	}
	"""Maps key error variables to their response error code"""

	__on_error = None
	"""On Error

	Function called when a service request raises an exception
	"""

	__service = ''
	"""The service's name"""

	__uris: dict = {}
	"""URIs
	used to keep track of the uris to callbacks for the purposes of __list calls
	"""

	__verbose = True
	"""The verbose mode, set to True to view requests/responses"""

	@classmethod
	def cors(cls, cors):
		"""CORs

		Sets the regular expression used to validate domains making requests

		Arguments:
			pattern (Re.Pattern): A compiled regular expression

		Returns:
			None
		"""
		cls.__cors = cors

	@classmethod
	def on_error(cls, callback: callable):
		"""On Error

		Sets the global error handling function

		Arguments:
			callback (callable): The function to call with the error details

		Returns:
			None
		"""
		cls.__on_error = staticmethod(callback)

	@classmethod
	def service(cls, s: str):
		"""Service

		Sets the name of the service associated with all routes

		Arguments:
			s (str): The name of the service

		Returns:
			None
		"""
		cls.__service = s

	@classmethod
	def verbose(cls, b: bool):
		"""Verbose

		Set the verbose mode associated with all routes

		Arguments:
			b (bool): True/False

		Returns:
			None
		"""
		cls.__verbose = b

	def __init__(self,
		callback: callable | Literal[True],
		uri: str | None = None
	):
		"""Constructor

		Initialises an instance of the route

		Arguments:
			callback (callable): The function to pass details to when this \
				route is triggered. Send True to make a __list route

		Returns:
			None
		"""

		# Store the callback
		self.__callback = callback

		# If we got a uri, associated the callback with it globally, this way
		#	we can reference it when getting __list requests
		if uri is not None:
			self.__uris[uri[1:]] = callback

	def __call__(self):
		"""Call (__call__)

		Python magic method that allows the instance to be called

		Returns:
			str
		"""

		# If CORS is enabled and the origin matches
		if self.__cors and \
			'origin' in bottle.request.headers and \
			self.__cors.match(bottle.request.headers['origin']):

			# Add the current origin as acceptable
			bottle.response.headers['Access-Control-Allow-Origin'] = \
				bottle.request.headers['origin']
			bottle.response.headers['Vary'] = 'Origin'

		# If the request is OPTIONS
		if bottle.request.method == 'OPTIONS':

			# If we're in verbose mode, let people know we are handling OPTIONS
			if self.__verbose:
				print('%s: %s OPTIONS %s' % (
					str(datetime.now()),
					self.__service,
					bottle.request.path
				))

			# Set the default headers expected for OPTIONS
			bottle.response.headers['Access-Control-Allow-Methods'] = \
				'DELETE, GET, POST, PUT, OPTIONS'
			bottle.response.headers['Access-Control-Max-Age'] = 1728000
			bottle.response.headers['Access-Control-Allow-Headers'] = \
				'Authorization,DNT,X-CustomHeader,Keep-Alive,User-Agent,' \
				'X-Requested-With,If-Modified-Since,Cache-Control,Content-Type'
			bottle.response.headers['Content-Type'] = 'text/plain charset=UTF-8'
			bottle.response.headers['Content-Length'] = 0
			bottle.request.status = 204
			return ''

		# Set the return to JSON
		bottle.response.headers['Content-Type'] = \
			'application/json; charset=utf-8'

		# Initialise the request details with the bottle request and response
		#	objects, allowing requests direct access to low level data
		oReq = jobject({
			'request': bottle.request,
			'response': bottle.response
		})

		# If we got a Read request and the data is in the GET
		if bottle.request.method == 'GET' and 'd' in bottle.request.query:

			# Convert the GET and store the data
			try:
				oReq.data = jsonb.decode(bottle.request.query['d'])
			except Exception as e:
				return response.Error((errors.REST_REQUEST_DATA, '%s\n%s' % (bottle.request.query['d'], str(e)))).to_json()

		# Else we most likely got the data in the body
		else:

			# Make sure the request send JSON
			try:
				if not self.__content_type.match(bottle.request.headers['Content-Type'].lower()):
					return str(response.Error(errors.REST_CONTENT_TYPE))
			except KeyError:
				return response.Error(errors.REST_CONTENT_TYPE).to_json()

			# Store the data, if it's too big we need to read it rather than
			#	use getvalue
			try:
				sData = bottle.request.body.getvalue()
			except AttributeError as e:
				sData = bottle.request.body.read()

			# Make sure we have a string, not a set of bytes
			try:
				sData = sData.decode()
			except (UnicodeDecodeError, AttributeError):
				pass

			# Convert the data and store it
			try:
				if sData: oReq.data = jsonb.decode(sData)
			except Exception as e:
				return response.Error(errors.REST_REQUEST_DATA,'%s\n%s' % (sData, str(e))).to_json()

		# If the request sent a authorization token
		if 'Authorization' in bottle.request.headers:

			# Get the session from the Authorization token
			oReq.session = memory.load(bottle.request.headers['Authorization'])

			# If the session is not found
			if not oReq.session:
				bottle.response.status = 401
				return response.Error(errors.REST_AUTHORIZATION, 'Unauthorized').to_json()

			# Else, extend the session's ttl
			else:
				oReq.session.extend()

		# In case the service crashes
		try:

			# If we're in verbose mode
			if self.__verbose:
				print('%s REQUEST %s %s %s %s' % (
					str(datetime.now()),
					self.__service,
					bottle.request.method,
					bottle.request.path,
					str('data' in oReq and jsonb.encode(oReq.data, 2) or 'None')
				))

			# If this is a list request
			if self.__callback is True:

				# If the data isn't passed or isn't an array
				if 'data' not in oReq or not isinstance(oReq.data, list):
					oResponse = response.Error(
						errors.REST_REQUEST_DATA,
						'data must be an array'
					)

				# Else, if it's beyond the max
				elif len(oReq.data) > 10:
					oResponse = response.Error(
						errors.REST_LIST_TO_LONG,
						'Can not request more than 10 urls via __list'
					)

				# Else, we have an array of nouns to call and return
				else:

					# Init the response
					lResponse = []

					# Go through each element in the list
					for m in oReq.data:

						# If we got a string
						if isinstance(m, str):
							m = [ m ]

						# Else, if we didn't get a list
						elif not isinstance(m, list):
							oResponse = response.Error(
								errors.REST_REQUEST_DATA, [ m,
									'data must be an array or URI and data, ' \
									'or single string for the URI' ]
							)
							break

						# If the URI doesn't exist
						if m[0] not in self.__uris:
							oResponse = response.Error(
								errors.REST_LIST_INVALID_URI,
								m[0]
							)
							break

						# Clone the __list request data so that the session
						#	or any other data is shared
						oRequest = clone(oReq)

						# If unique data was passed for the child request
						if len(m) == 2:

							# If we didn't get a dict
							if not isinstance(m[1], dict):
								oResponse = response.Error(
									errors.REST_LIST_INVALID_URI,
									[ m[1], 'data must be an object' ]
								)
								break

							# Set the data for this request
							oRequest.data = m[1]

						# No unique data was sent, so we need to delete the
						#	original data from the request
						else:
							try: del oRequest.data
							except AttributeError: pass

						# Call the request and append the data to the
						#	response
						try:
							lResponse.append([
								m[0],
								self.__uris[m[0]](oRequest).to_dict()
							])

						# If we got a KeyError
						except (AttributeError, KeyError) as e:
							if e.args[0] in self.__key_to_errors:
								oResponse = response.Error(
									self.__key_to_errors[e.args[0]]
								)
								break
							else:
								raise

						# If we got a response exception
						except response.ResponseException as e:

							# Set the response using the exceptions first argument
							oResponse = e.args[0]
							break

					# Else, we got through the list
					else:

						# Set the response using the list of individual
						#	responses
						oResponse = response.Response(lResponse)

			# Else, we are making a single URI request
			else:

				# Call the appropriate API method based on the HTTP/request
				#	method
				try:
					oResponse = self.__callback(oReq)

				# If we got a KeyError
				except (AttributeError, KeyError) as e:
					if e.args[0] in self.__key_to_errors:
						oResponse = response.Error(
							self.__key_to_errors[e.args[0]]
						)
					else:
						raise

				# If we got a response exception
				except response.ResponseException as e:

					# Set the response using the exceptions first argument
					oResponse = e.args[0]

		# If we get absolutely any exception
		except Exception as e:

			# Get the traceback info
			sError = traceback.format_exc()

			# Print the traceback to stderr
			print(sError, file=sys.stderr)

			# If we have an error handler
			if self.__on_error:

				# Gather all the details, including optional ones
				oDetails = {
					'service': self.__service,
					'method': bottle.request.method,
					'path': bottle.request.path,
					'environment': bottle.request.environ,
					'traceback': sError
				}
				for s in ['data', 'session']:
					if s in oReq:
						oDetails[s] = oReq[s]

				# Pass the details to the error handler
				self.__on_error(oDetails)

			# Set a response of service/request crashed
			oResponse = response.Error(
				errors.SERVICE_CRASHED,
				'%s:%s' % (self.__service, bottle.request.path)
			)

		# If the response contains an error
		if oResponse.error:

			# If it's an authorization error
			if oResponse.error['code'] == errors.REST_AUTHORIZATION:

				# Set the http status to 401 Unauthorized
				bottle.response.status = 401

				# If the message is missing
				if oResponse.error['msg'] == '':
					oResponse.error['msg'] = 'Unauthorized'

			# Add the service and path to the call
			try:
				oResponse.error['service'].append([
					self.__service, bottle.request.method, bottle.request.path
				])
			except KeyError:
				oResponse.error['service'] = [
					[self.__service, bottle.request.method, bottle.request.path]
				]

		# If we're in verbose mode
		if self.__verbose:
			print('%s RETURNING %s %s %s %s' % (
				str(datetime.now()),
				self.__service,
				bottle.request.method,
				bottle.request.path,
				jsonb.encode(oResponse.to_dict(), 2)
			)
		)

		# Return the Response as a string
		return oResponse.to_json()

class REST(bottle.Bottle):
	"""REST

	Used to access a service instance via HTTP/REST

	Extends:
		Bottle
	"""

	__action_to_method = {
		'create': 'POST',
		'delete': 'DELETE',
		'read':   'GET',
		'update': 'PUT'
	}
	"""Maps HTTP methods to service actions"""

	__noun_regex = re.compile(
		r'([a-z]+(?:_[a-z]+)*)_(create|delete|read|update)'
	)
	"""Regular Expression to match to valid service noun method"""

	def __init__(self,
		name: str,
		instance: service.Service,
		cors: List[str] = None,
		lists = True,
		on_errors: callable = None,
		verbose: bool = False
	):
		"""Constructor

		Creates a new REST instance

		Arguments:
			name (str): The name of the service
			instance (Service): The service to make accessible via REST
			cors (str[]): A list of allowed domains for CORS policy
			lists (bool | dict): True to add a __list call to the default
				service, else a dictionay of uris to services
			on_errors (callable): Optional, a function to call when a service \
				request throws an exception

		Raises:
			ValueError

		Returns:
			RestService
		"""

		# Call the parent constructor first so the object is setup
		super(REST, self).__init__()

		# If the instance is not a Service
		if not isinstance(instance, service.Service):
			raise ValueError('Invalid service passed to %s' %
								sys._getframe().f_code.co_name)

		# Store the service
		self.instance = instance

		# If cors, compile it
		if cors:

			# If it's not a list
			if not isinstance(cors, list):
				raise RuntimeError('REST.cors must be a list')

			# If we only have one
			if len(cors) == 1:
				cors = cors[0].replace('.', '\\.')
			else:
				cors = '(?:%s)' % '|'.join([
					s.replace('.', '\\.')
					for s in cors
				])
			cors = re.compile('https?://(.*\\.)?%s' % cors)

			# Set it
			_Route.cors(cors)

		# Set the service name
		_Route.service(name)

		# If we have an error handler
		if on_errors:
			_Route.on_error(on_errors)

		# Set the verbose mode
		_Route.verbose(verbose)

		# Go through all the functions found on the service
		for sFunc in dir(self.instance):

			# Check the format of the method name
			oMatch = self.__noun_regex.match(sFunc)

			# If it's a match
			if oMatch:

				# Get the method
				sMethod = self.__action_to_method[oMatch.group(2)]

				# Generate the URL
				sUri = '/' + ('/'.join(oMatch.group(1).split('_')))

				# Register it with bottle
				self.route(
					sUri,
					[ sMethod, 'OPTIONS' ],
					_Route(
						getattr(self.instance, sFunc),	# The function to call
						uri = (lists and sMethod == 'GET') and sUri or None
					)
				)

		# If we have a request for a list of requests
		if lists:

			# If it's True
			if lists is True:
				lists = { '/__list': service }

			# Go through each request in the list
			for sUri in lists:

				# Add the list read route
				self.route(
					sUri,
					[ 'GET', 'OPTIONS' ],
					_Route(True)
		)

	# run method
	def run(self, server='gunicorn', host='127.0.0.1', port=8080,
			reloader=False, interval=1, quiet=False, plugins=None,
			debug=None, maxfile=20971520, **kargs):
		"""Run

		Overrides Bottle's run to default gunicorn and other fields

		Arguments:
			server (str): Server adapter to use
			host (str): Server address to bind to
			port (int): Server port to bind to
			reloader (bool): Start auto-reloading server?
			interval (int): Auto-reloader interval in seconds
			quiet (bool): Suppress output to stdout and stderr?
			plugins (list): List of plugins to the server
			debug (bool): Debug mode
			maxfile (int): Maximum size of requests

		Returns:
			None
		"""

		# Set the max file size
		bottle.BaseRequest.MEMFILE_MAX = maxfile

		# Call bottle run
		bottle.run(
			app=self, server=server, host=host, port=port, reloader=reloader,
			interval=interval, quiet=quiet, plugins=plugins, debug=debug,
			**kargs
		)


# coding=utf8
""" Service

Holds the class used to create services that can be started as rest apps
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-15"

# Ouroboros imports
from tools import evaluate

# Python imports
import abc

# Module imports
from body.errors import DATA_FIELDS
from body.response import ResponseException

class Service(abc.ABC):
	"""Service

	The object to build all services from
	"""

	@staticmethod
	def check_data(data: dict, fields: list):
		"""Check Data

		Checks if `fields` are set in the `data` dictionary. Raises a \
		DATA_FIELDS ResponseException if any of the `fields` or sub-fields are \
		missing

		\# Check '_id' and 'name' exist in req.data

		check_data(req.data, [ '_id', 'name' ])

		\# Check '_id' and 'record.name' exist in req.data

		check_data(req.data, [ '_id', { 'record': [ 'name' ]} ])

		\# Check 'record.name', and 'options.raw' exist in req.data

		check_data(req.data, {
			'record': [ 'name' ],
			'options': [ 'raw' ]
		})

		Arguments:
			data (dict): The dict to check for missing fields
			fields (list | dict): The list of fields to check for

		Raises:
			ResponseException
		"""

		# Check the data
		try:
			evaluate(data, fields)
		except ValueError as e:
			raise ResponseException(error = (
				DATA_FIELDS,
				[ [ f, 'missing' ] for f in e.args ]
			))

	@abc.abstractmethod
	def reset(self):
		"""Reset

		Called when the system has been reset, usually by loading new data that
		the instance will need to process/reprocess

		Returns:
			None
		"""
		raise NotImplementedError('Must implement the "reset" method')


# coding=utf8
"""Body

Shared methods for accessing the brain and other shared formats
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2022-08-29"

__all__ = [
	'constants', 'Error', 'errors', 'regex', 'register_services', 'Response',
	'ResponseException', 'REST', 'Service'
]

# Ouroboros imports
from config import config
import jsonb
import undefined

# Python imports
from copy import copy
from sys import stderr
from time import sleep

# Pip imports
import requests

# Local imports
from . import constants, errors, locations, regex, response, rest, service

# Error, Response, and ResponseException
Error = response.Error
Response = response.Response
ResponseException = response.ResponseException

# REST and Config
REST = rest.REST
Locations = locations.Locations

# Service
Service = service.Service

__services = {}
"""Registered Services"""

__action_to_request = {
	'create': [ requests.post, 'POST' ],
	'delete': [ requests.delete, 'DELETE' ],
	'read': [ requests.get, 'GET' ],
	'update': [ requests.put, 'PUT' ]
}
"""Map actions to request methods"""

def create(
	service: str,
	path: str,
	req: dict = {},
	headers: dict = undefined
):
	"""Create

	Make a POST request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'
		headers (dict): Any additional headers to add to the request

	Returns:
		Response
	"""
	return request(service, 'create', path, req, headers)

def delete(
	service: str,
	path: str,
	req: dict = {},
	headers: dict = undefined
):
	"""Delete

	Make a DELETE request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'
		headers (dict): Any additional headers to add to the request

	Returns:
		Response
	"""
	return request(service, 'delete', path, req, headers)

def read(
	service: str,
	path: str,
	req: dict = {},
	headers: dict = undefined
):
	"""Read

	Make a GET request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'
		headers (dict): Any additional headers to add to the request

	Returns:
		Response
	"""
	return request(service, 'read', path, req, headers)

def register_services(running: dict = undefined) -> Locations:
	"""Register Services

	Takes a dictionary of services to their urls for use by the request
	functions

	Arguments:
		running (dict): Optional, dict of local running services

	Returns:
		body.locations.Locations
	"""

	# Pull in the global services
	global __services

	# Init the REST config
	oLocations = Locations(config.body.rest({
		'allowed': None,
		'default': {
			'domain': 'localhost',
			'host': '0.0.0.0',
			'port': 9000,
			'protocol': 'http',
			'workers': 1
		}
	}))

	# Rest the dict
	__services = {}

	# Go through each config
	for s in oLocations:

		# If the service exists locally
		if running and s in running:
			__services[s] = running[s]

		# Else, add the URL
		else:
			__services[s] = oLocations[s]['url']

	# Return the config
	return oLocations

def request(
	service: str,
	action: str,
	path: str,
	req: dict = {},
	headers: dict = undefined
):
	"""Request

	Method to convert REST requests into HTTP requests

	Arguments:
		service (str): The service we are requesting data from
		action (str): The action to take on the service
		path (str): The path of the request
		req (dict): The request details: 'data', 'session', and 'enviroment'
		headers (dict): Any additional headers to add to the request

	Raises:
		KeyError: if the service or action don't exist

	Return:
		Response
	"""

	# Init the data and headers
	sData = ''
	dHeaders = (headers is not undefined and isinstance(headers, dict)) and \
				copy(headers) or \
				{}

	# Add the default content length and type
	dHeaders['Content-Length'] = '0'
	dHeaders['Content-Type'] = 'application/json; charset=utf-8'

	# If the data was passed
	if 'data' in req and req['data']:

		# Convert the data to JSON and store the length
		sData = jsonb.encode(req['data'])
		dHeaders['Content-Length'] = str(len(sData))

	# If we have a session, add the ID to the headers
	if 'session' in req and req['session']:
		dHeaders['Authorization'] = req['session'].key()

	# Loop requests so we don't fail just because of a network glitch
	iAttempts = 0
	while True:

		# Increase the attempts
		iAttempts += 1

		# If we got a service instance
		if isinstance(__services[service], Service):
			print('Can not currently call local services', file = stderr)

		# Else, this is an external service
		else:

			# Make the request using the services URL and the current path, then
			#	store the response
			try:
				oRes = __action_to_request[action][0](
					__services[service] + path,
					data=sData,
					headers=dHeaders
				)

				# If the request wasn't successful
				if oRes.status_code != 200:

					# If we got a 401
					if oRes.status_code == 401:
						return Response.from_json(oRes.content)
					else:
						return Error(
							errors.SERVICE_STATUS,
							'%d: %s' % (oRes.status_code, oRes.content)
						)

				# If we got the wrong content type
				if oRes.headers['Content-Type'].lower() != \
					'application/json; charset=utf-8':
					return Error(
						errors.SERVICE_CONTENT_TYPE,
						'%s' % oRes.headers['content-type']
					)

				# Success, break out of the loop
				break

			# If we couldn't connect to the service
			except requests.ConnectionError as e:

				# If we haven't exhausted attempts
				if iAttempts < 3:

					# Wait for a second
					sleep(1)

					# Loop back around
					continue

				# We've tried enough, return an error
				return Error(errors.SERVICE_UNREACHABLE, str(e))

	# Else turn the content into a Response and return it
	return Response.from_json(oRes.text)

def update(
	service: str,
	path: str,
	req: dict = {},
	headers: dict = undefined
):
	"""Update

	Make a PUT request

	Arguments:
		service (str): The service to call
		path (str): The path on the service
		req (dict): The request details, which can include 'data' and 'session'
		headers (dict): Any additional headers to add to the request

	Returns:
		Response
	"""
	return request(service, 'update', path, req, headers)


# coding=utf8
"""Locations

Configuration for accessing REST services via http
"""

__author__		= "Chris Nasr"
__copyright__	= "Ouroboros Coding Inc."
__email__		= "chris@ouroboroscoding.com"
__created__		= "2023-03-17"

class Locations(object):
	"""Locations class

	Represents configuration data for connecting to/loading services
	"""

	def __contains__(self, service):
		"""__contains__

		Python magic method for checking a key exists in a dict like object

		Arguments:
			service (str): The service to check for

		Returns:
			bool
		"""
		return service in self.__services

	def __getitem__(self, service):
		"""__getitem__

		Python magic method for getting a key from a dict like object

		Arguments:
			service (str): The service config to return

		Raises:
			KeyError

		Returns:
			mixed
		"""

		# If it's in services
		try:
			return self.__services[service].copy()

		# Else, throw an exception
		except KeyError:
			raise KeyError(service)

	def __init__(self, conf):
		"""Constructor

		Initialises the instance

		Arguments:
			conf (dict): The configuration data for compiling the list of
				services

		Returns:
			Locations
		"""

		# If we didn't get a dictionary for the service conf
		if not isinstance(conf, dict):
			raise ValueError('conf')

		# If we didn't get a list of services
		if 'services' not in conf:
			raise ValueError('conf.services')

		# Init the defaults if none are found
		if 'defaults' not in conf:
			conf['defaults'] = {}

		# Port values are not modified by default
		iPortMod = 0

		# If there is a port modifier
		if 'port' in conf['default']:

			# Make sure it's an integer
			try:
				iPortMod = int(conf['default']['port'])
				del conf['default']['port']
			except ValueError:
				raise ValueError('conf.default.port must be an int')

		# Initialise the list of services
		self.__services = {}

		# Loop through the list of services
		for s in conf['services']:

			# If the service doesn't point to a dict
			if not isinstance(conf['services'][s], dict):
				raise ValueError('conf.services.%s' % s)

			# Start with the default values
			dParts = conf['default'].copy()

			# Then add the service values
			dParts.update(conf['services'][s])

			# If we have no port
			if 'port' not in dParts:

				# But we have a modifier, assume we add to 80
				if iPortMod: dParts['port'] = 80 + iPortMod

			# Else add the modifier to the port passed
			else:
				dParts['port'] += iPortMod

			# Set defaults for any missing parts
			if not dParts['protocol']: dParts['protocol'] = 'http'
			if not dParts['domain']: dParts['domain'] = 'localhost'
			if 'path' not in dParts: dParts['path'] = ''
			else: dParts['path'] = '%s/' % str(dParts['path'])

			# Store the parts for the service
			self.__services[s] = dParts.copy()

			# Generate a URL from the parts and store it
			self.__services[s]['url'] = '%s://%s%s/%s' % (
				dParts['protocol'],
				dParts['domain'],
				'port' in dParts and ":%d" % dParts['port'] or '',
				dParts['path']
			)

			# If we still have no port, default to 80
			if 'port' not in self.__services[s]:
				self.__services[s]['port'] = 80

	def __iter__(self):
		"""__iter__

		Python magic method to return an iterator for the instance

		Returns:
			iterator
		"""
		return iter(self.__services)

	def __str__(self):
		"""__str__

		Python magic method to return a string for the instance

		Returns:
			str
		"""
		return str(self.__services)

	def keys(self):
		"""services

		Returns the keys (services) in the instance

		Returns:
			str[]
		"""
		return self.__services.keys()


from setuptools import setup

with open('README.md', 'r') as oF:
	long_description=oF.read()

setup(
	name='body_oc',
	version='2.0.3',
	description='Body contains shared concepts among all body parts',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url='https://ouroboroscoding.com/body/',
	project_urls={
		'Documentation': 'https://ouroboroscoding.com/body/',
		'Source': 'https://github.com/ouroboroscoding/body',
		'Tracker': 'https://github.com/ouroboroscoding/body/issues'
	},
	keywords=[ 'rest', 'microservices' ],
	author='Chris Nasr - Ouroboros Coding Inc.',
	author_email='chris@ouroboroscoding.com',
	license='Custom',
	packages=[ 'body' ],
	package_data={ 'body': [
		'docs/*',
		'docs/templates/*.j2'
	] },
	python_requires='>=3.10',
	install_requires=[
		'bottle>=0.12.23,<0.13',
		'gunicorn>=21.2.0,<21.3',
		'jobject>=1.0.3,<1.1.0',
		'jsonb>=1.0.0,<1.1.0',
		'memory-oc>=1.0.0,<1.1',
		'strings-oc>=1.0.7,<1.1',
		'requests>=2.31.0,<2.32',
		'undefined-oc>=1.0.0,<1.1',
		'tools-oc>=1.2.5,<1.3'
	],
	entry_points={
		'console_scripts': ['body-docs=body.docs.__main__:cli']
	},
	zip_safe=True
)